//
//	ADUtilities.m
//	  Written by Carl Lindberg Copyright 1998 by Carl Lindberg.
//                     All rights reserved.
//		This notice may not be removed from this source code.
//
//	This header is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"License.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

#import "ADUtilities.h"
#import "ADDefaults.h"
#import <stdio.h>
#import <Foundation/NSPathUtilities.h>
#import <Foundation/NSCharacterSet.h>
#import <Foundation/NSUtilities.h>
#import <Foundation/NSDate.h>
#import <Foundation/NSBundle.h>
#import <Foundation/NSArray.h>
#if HAVE_NSCALENDARDATE_H
#import <Foundation/NSCalendarDate.h>
#endif
#import "NSFileManager+Autodoc.h"
#import <MiscFoundation.subproj/NSString+MiscRegex.h>
#import <MiscFoundation.subproj/NSString+MiscAdditions.h>
#import "ADOutputModule.h"

/* Logs to stdout */
void ADLogv(NSString *format, va_list arguments)
{
    NSString *logString = [[NSString alloc] initWithFormat:format arguments:arguments];

    fwrite([logString cString], 1, [logString cStringLength], stdout);
    if (![logString hasSuffix:@"\n"]) fputc('\n', stdout); //hrm

    [logString release];
}

void ADLog(NSString *format, ...)
{
    va_list arguments;

    va_start(arguments, format);
    ADLogv(format, arguments);
    va_end(arguments);
}

void ADDebugLog(int debugLevel, NSString *format, ...)
{
    if (debugLevel <= ADIntForDefault(ADDebugLevel))
    {
        va_list arguments;

        va_start(arguments, format);
        ADLogv(format, arguments);
        va_end(arguments);
    }
}

/*
 * Our Flex/bison files "#define fprintf adfprintf", with the result being
 * that I can direct everything to stdout.
 */
void adfprintf(FILE *fp, const char *format, ...)
{
    va_list arguments;

    va_start(arguments, format);
    vfprintf(stdout, format, arguments);
    va_end(arguments);
}

#define ADD_ARRAY(array)             \
{                                    \
    NSArray *theArray = array;       \
    int i, count = [theArray count]; \
    for (i=0; i<count; i++)          \
    {                                \
        [searchPath addObject:[[theArray objectAtIndex:i] stringByAppendingPathComponent:@"AutoDoc"]]; \
    }                                \
}

NSArray *ADTemplateSearchPath()
{
    NSMutableArray *searchPath = [NSMutableArray array];
#if defined(YELLOWBOX) || defined(GNUSTEP)
    // gnustep-base 0.6.6 raises exception if GNUSTEP_LOCAL_ROOT not defined...
    int locationMask = NSUserDomainMask|NSLocalDomainMask|NSNetworkDomainMask;
    ADD_ARRAY(NSSearchPathForDirectoriesInDomains(NSAllLibrariesDirectory, locationMask, YES));
#else
    // How to add ~/Developer/AutoDoc, /LocalDeveloper/AutoDoc ?
    ADD_ARRAY(NSStandardLibraryPaths());
#endif
    [searchPath addObject:[[NSBundle mainBundle] bundlePath]];

    return searchPath;
}

NSCharacterSet *ADVariableNameSet()
{
    static NSMutableCharacterSet *variableSet = nil;

    if (!variableSet)
    {
        variableSet = [[NSCharacterSet alphanumericCharacterSet] mutableCopy];
        [variableSet addCharactersInString:@"_"];
    }

    return variableSet;
}

NSCharacterSet *ADNonVariableNameSet()
{
    static NSMutableCharacterSet *nonVariableSet = nil;

    if (!nonVariableSet)
    {
        nonVariableSet = [[ADVariableNameSet() invertedSet] retain];
    }

    return nonVariableSet;
}

BOOL ADIsCopyrightComment(NSString *commentString)
{
    return [commentString containsRegex:@"[$](Id|Header|Date|Revision):"];
}

BOOL ADIsAutodocComment(NSString *commentString)
{
    if ([commentString hasPrefix:@"//"])
    {
        return ADBoolForDefault(ADAcceptDoubleSlashComments);
    }

    /* Okay, it's a C-style comment */

    // If it's a comment like /****************/, ignore it.
    if ([commentString containsRegex:@"\\`/[*]+/\\'"])
    {
        return NO;
    }

    /* If we're accepting all C-style comments, allow anything at this point. */
    if (ADBoolForDefault(ADAcceptStandardCComments))
        return YES;

    /* Autodoc actually allows comments like /****"  */
    return ([commentString containsRegex:@"\\`/[*]+\""] &&
            [commentString containsRegex:@"\"[*]+/\\'"]);    
}

NSRange ADRangeOfMatchedLeftString(NSString *aString, NSString *leftChar, NSString *rightChar)
{
    NSRange searchRange = {0, [aString length] };
    NSString *combinedString = [leftChar stringByAppendingString:rightChar];
    NSCharacterSet *parenSet = [NSCharacterSet characterSetWithCharactersInString:combinedString];
    int matchCount = 0;
    NSRange foundRange;
    NSString *foundString;

    while (1)
    {
        foundRange = [aString rangeOfCharacterFromSet:parenSet
                                              options:NSBackwardsSearch
                                                range:searchRange];

        if (foundRange.length == 0) return foundRange;

        foundString = [aString substringWithRange:foundRange];

        if ([foundString isEqualToString:rightChar]) matchCount++;
        else matchCount--;

        if (matchCount == 0) return foundRange;

        searchRange.length = foundRange.location;
    }

    return NSMakeRange(0,0);
}

NSRange ADRangeOfMatchedLeftParen(NSString *aString)
{
    return ADRangeOfMatchedLeftString(aString, @"(", @")");
}

NSRange ADRangeOfMatchedLeftSquareBracket(NSString *aString)
{
    return ADRangeOfMatchedLeftString(aString, @"[", @"]");
}

NSRange ADRangeOfMatchedRightBrace(NSString *aString, int startIndex)
{
    NSRange searchRange = {startIndex, [aString length] - startIndex };
    NSCharacterSet *delimSet = [NSCharacterSet characterSetWithCharactersInString:@"{}"];
    int matchCount = 0;
    NSRange foundRange;
    NSString *foundString;

    while (1)
    {
        foundRange = [aString rangeOfCharacterFromSet:delimSet options:0 range:searchRange];

        if (foundRange.length == 0) return foundRange;

        foundString = [aString substringWithRange:foundRange];
        if ([foundString isEqualToString:@"}"]) matchCount--;
        else matchCount++;

        if (matchCount == 0) return foundRange;

        searchRange.location = NSMaxRange(foundRange);
        searchRange.length = [aString length] - searchRange.location;
    }

    return NSMakeRange(0,0);
}

NSRange ADRangeOfVariableBeforeIndex(NSString *string, unsigned lastIndex)
{
    NSRange searchRange = {0, lastIndex};
    NSRange nameStartRange, nameEndRange;
    NSRange variableRange;

    nameEndRange = [string rangeOfCharacterFromSet:ADVariableNameSet()
                                           options:NSBackwardsSearch
                                             range:searchRange];

    if (nameEndRange.length == 0) return NSMakeRange(NSNotFound, 0);

    searchRange.length = nameEndRange.location;
    nameStartRange = [string rangeOfCharacterFromSet:ADNonVariableNameSet()
                                             options:NSBackwardsSearch
                                               range:searchRange];

    if (nameStartRange.length == 0)
        variableRange.location = 0;
    else
        variableRange.location = NSMaxRange(nameStartRange);
    variableRange.length = NSMaxRange(nameEndRange) - variableRange.location;

    return variableRange;
}

/*
 * Used by typedefs, ivars, and global vars; this attempts to pluck the
 * variable name out of the fully-typed variable string.  This is generally
 * the last word in the string, but this skips over type information that
 * can follow the variable name (currently bitfields, arrays, and function
 * pointer definition stuff).
 */
NSRange ADRangeOfVariableInTypeString(NSString *string)
{
    NSString *varString   = [string stringByTrimmingTailWhitespace];
    NSRange bitfieldRange = [varString rangeOfRegex:@":[0-9]+[ \t\n\r]*;\\'"];
    NSRange arrayRange    = [varString rangeOfRegex:@"([ \t\n\r]*[[][^]]*[]])+[ \t\n\r]*;\\'"];
    NSRange funcRange     = [varString rangeOfRegex:@"[)][ \t\n\r]*;\\'"];

    if (bitfieldRange.location != NSNotFound)
    {		
        return ADRangeOfVariableBeforeIndex(varString, bitfieldRange.location);
    }
    else if (arrayRange.location != NSNotFound)
    {
        return ADRangeOfVariableBeforeIndex(varString, arrayRange.location);
    }
    else if (funcRange.location != NSNotFound) // typed function ivar
    {
        NSRange leftParenRange  = ADRangeOfMatchedLeftParen(varString);
        return ADRangeOfVariableBeforeIndex(varString, leftParenRange.location);
    }
    else
    {
        return ADRangeOfVariableBeforeIndex(varString, [varString length]);
    }
}

NSString *ADSquashTypeString(NSString *aString)
{
    NSCharacterSet *charSet;

    /* If it's a struct or union type, leave newlines alone. */
    if ([aString containsString:@"{"])
        charSet = [NSCharacterSet characterSetWithCharactersInString:@" \t\v\f"];
    else
        charSet = [NSCharacterSet characterSetWithCharactersInString:@" \t\n\r\v\f"];

    aString = [aString stringByTrimmingLeadWhitespace];
    aString = [aString stringByReplacingEverySeriesOfCharactersFromSet:charSet withString:@" "];

    return aString;
}

/* Could be an NSString method */
NSString *ADPathWithExtension(NSString *path, NSString *extension)
{
    return [[path stringByDeletingPathExtension] stringByAppendingPathExtension:extension];
}

NSString *ADNormalizePath(NSString *path)
{
    NSFileManager *manager = [NSFileManager defaultManager];
    NSString      *fullPath = path;

    if (![fullPath isAbsolutePath])
        fullPath = [[manager currentDirectoryPath] stringByAppendingPathComponent:fullPath];

    fullPath = [fullPath stringByStandardizingPath];

    /* in case stringByStandardizingPath returns nil */
    if ([fullPath length] == 0) fullPath = path;

    return fullPath;
}

BOOL ADFileExists(NSString *filename)
{
    return [[NSFileManager defaultManager] regularFileExistsAtPath:filename];
}

BOOL ADFileIsNewer(NSString *file1, NSString *file2)
{
    NSFileManager *manager = [NSFileManager defaultManager];
    NSDate *date1 = [[manager fileAttributesAtPath:file1 traverseLink:YES] fileModificationDate];
    NSDate *date2 = [[manager fileAttributesAtPath:file2 traverseLink:YES] fileModificationDate];

    if (date1 == nil || date2 == nil) return YES;
    return ([date1 compare:date2] == NSOrderedDescending);
}

/* May want a way to turn auto-bolding off... */
NSArray *ADBoldWordsArray()
{
    static NSArray *boldArray = nil;

    if (boldArray == nil)
    {
        boldArray = [[NSArray alloc] initWithObjects:@"id", @"self", @"super", @"nil", nil];
    }

    return boldArray;
}

/*
 * Separate function to bold or italicize certain words.  We need this in
 * its own function so the fixed width [ !{...} markup ] formatting
 * function can call us selectively -- we don't want to auto-bold words in
 * fixed-width sections.
 */
static NSString *ADMarkupSpecialWords(NSString *string, NSArray *italicWords)
{
    ADOutputModule *module = [ADOutputModule currentModule];
    NSString *replaceString;
    NSArray  *boldWords = ADBoldWordsArray();
    int      i, count;

    /* Make double newline markers into paragraph separators.  This
        is in here because we don't want to do this for fixed-width sections. */
    string = [string stringByReplacingEveryOccurrenceOfRegex:@"\r?\n\r?\n[\n\r\t ]*"
                                                  withString:[module paragraphSeparator]];

    /* boldify words in boldWordsArray */
    replaceString  = [NSString stringWithFormat:@"$1%@$2%@$3",
                         [module startBold], [module stopBold]];
    for (i=0, count=[boldWords count]; i<count; i++)
    {
        NSString *currWord = [boldWords objectAtIndex:i];
        NSString *regex = [NSString stringWithFormat:
                            @"(\\`|[^_a-zA-Z0-9])(%@)(\\'|[^_a-zA-Z0-9])", currWord];

        string = [string stringByReplacingEveryOccurrenceOfRegex:regex
                            withString:replaceString
                            options:MiscUseMatchSubstitutions];
    }

    /* italicize words in italicWordsArray */
    replaceString  = [NSString stringWithFormat:@"$1%@$2%@$3",
                         [module startItalics], [module stopItalics]];
    for (i=0, count=[italicWords count]; i<count; i++)
    {
        NSString *currWord = [italicWords objectAtIndex:i];
        NSString *regex = [NSString stringWithFormat:
                            @"(\\`|[^_a-zA-Z0-9])(%@)(\\'|[^_a-zA-Z0-9])", currWord];

        if ([currWord length] == 0) continue;   //Make sure
        string = [string stringByReplacingEveryOccurrenceOfRegex:regex
                            withString:replaceString
                            options:MiscUseMatchSubstitutions];
    }

    return string;
}

static NSString *ADFormatValue(NSString *string, BOOL allowGenericProcessing)
{
    ADOutputModule *outputModule = [ADOutputModule currentModule];
    NSString       *indentString = @"__INDENT__";

    string = [string stringByReplacingEveryOccurrenceOfRegex:@"\r?\n[\t ]*" withString:@"\n"];
    string = [string stringByReplacingEveryOccurrenceOfRegex:@"[\t ]+" withString:@" "];

    if ([string containsString:@"{"] && [string containsString:@"\n"])
    {
        NSString *indent = @"";
        NSArray *lineArray = [string componentsSeparatedByString:@"\n"];
        NSEnumerator *lineEnum = [lineArray objectEnumerator];
        NSString *line;
        NSMutableArray *newArray = [NSMutableArray arrayWithCapacity:[lineArray count]];

        while (line = [lineEnum nextObject])
        {
            if ([line containsString:@"{"])
            {
                NSString *replace;

                replace = [NSString stringWithFormat:@"{%@$1", indentString];
                line = [line stringByReplacingRegex:@"{ *([^ ]*)"
                            withString:replace
                            options:MiscUseMatchSubstitutions];
                line = [indent stringByAppendingString:line];
                indent = [indent stringByAppendingString:indentString];
            }
            else if ([line containsString:@"}"])
            {
                if ([indent length] >= [indentString length])
                    indent = [indent substringFromIndex:[indentString length]];
                line = [indent stringByAppendingString:line];
            }
            else
            {
                line = [indent stringByAppendingString:line];
            }

            [newArray addObject:line];
        }

        string = [newArray componentsJoinedByString:@"\n"];
    }

    if (allowGenericProcessing)
        string = [outputModule preProcessString:string];
    string = [string stringByReplacingEveryOccurrenceOfString:@"\n"
                    withString:[outputModule lineSeparator]];
    string = [string stringByReplacingEveryOccurrenceOfString:indentString
                    withString:[outputModule valueIndentString]];
    if (allowGenericProcessing)
        string = [outputModule postProcessString:string];

    return string;
}

/* This is used to indent typedefs and other 'type' values */
NSString *ADFormatValueString(NSString *string)
{
    return ADFormatValue(string, YES);
}


/* Does the formatting for _{...} type of markup. */
static NSString *ADFormatListMarkup(NSString *string, ADOutputModule *outputModule)
{
    NSMutableString *newString = [NSMutableString stringWithCapacity:[string length]];
    unsigned stringLength = [string length];
    NSRange searchRange = NSMakeRange(0, stringLength);
    NSRange foundRange;
    NSRange betweenRange = NSMakeRange(0, 0);
    NSString *listRegex = @"_{[ \n\t\r]*((\\\\}|[^} \n\t\r])+)[ \n\t\r]*((\\\\}|[^}])*)} *";
    NSString *listGroupRegex = [NSString stringWithFormat:@"(%@[ \n\r\t]*)+", listRegex];

    foundRange = [string rangeOfRegex:listGroupRegex range:searchRange];
    while (foundRange.location != NSNotFound)
    {
        NSString *listsString;
        betweenRange.length = foundRange.location - betweenRange.location;
        [newString appendString:[string substringWithRange:betweenRange]];

        [newString appendString:[outputModule listPrefixString]];
        listsString = [string substringWithRange:foundRange];
        listsString = [listsString stringByReplacingEveryOccurrenceOfRegex:listRegex
                        withString:[outputModule listItemReplaceString]
                        options:MiscUseMatchSubstitutions];
        listsString = [listsString stringByReplacingEveryOccurrenceOfString:@"\\}"
                        withString:[outputModule preProcessString:@"}"]];

        /* Reformat paragraph separators with special list versions */
        listsString = [listsString stringByReplacingEveryOccurrenceOfString:
                        [outputModule paragraphSeparator]
                        withString:[outputModule listParagraphSeparator]];

        [newString appendString:listsString];
        [newString appendString:[outputModule listPostfixString]];

        betweenRange.location = NSMaxRange(foundRange);
        searchRange = NSMakeRange(betweenRange.location, stringLength - betweenRange.location);
        foundRange = [string rangeOfRegex:listGroupRegex range:searchRange];
    }

    [newString appendString:[string substringFromIndex:betweenRange.location]];

    return newString;
}

/* Does the formatting for !{...} type of markup. */
static NSString *ADFormatFixedFontMarkup(NSString *string, NSArray *italicWords)
{
    ADOutputModule  *module = [ADOutputModule currentModule];
    NSMutableString *newString = [NSMutableString stringWithCapacity:[string length]];
    unsigned stringLength = [string length];
    NSRange searchRange = NSMakeRange(0, stringLength);
    NSRange foundRange;
    NSRange betweenRange = NSMakeRange(0, 0);

    foundRange = [string rangeOfString:@"!{" options:0 range:searchRange];

    while (foundRange.length != 0)
    {
        NSRange rightRange = ADRangeOfMatchedRightBrace(string, foundRange.location);
        NSRange blockRange;
        NSString *blockString;

        if (rightRange.length == 0) break;

        /*
         * Append the regular text before the !{.  We only do the
         * autobolding etc. on this string, since we want to avoid that in
         * the fixed-width stuff.
         */
        betweenRange.length = foundRange.location - betweenRange.location;
        [newString appendString:
            ADMarkupSpecialWords([string substringWithRange:betweenRange], italicWords)];

        blockRange.location = NSMaxRange(foundRange);
        blockRange.length = rightRange.location - blockRange.location;
        blockString = [string substringWithRange:blockRange];

        [newString appendString:[module startFixedFont]];
        blockString = ADFormatValue(blockString, NO);
        if ([blockString containsString:@"\n"])
            blockString = ADIndentString(blockString, 1, YES);
        [newString appendString:blockString];
        [newString appendString:[module stopFixedFont]];

        betweenRange.location = NSMaxRange(rightRange);
        searchRange = NSMakeRange(betweenRange.location, stringLength - betweenRange.location);
        foundRange = [string rangeOfString:@"!{" options:0 range:searchRange];
    }

    [newString appendString:
        ADMarkupSpecialWords([string substringFromIndex:betweenRange.location], italicWords)];

    return newString;
}

/*"
 * Formats the string, converting autodoc-style markup into the appropriate
 * output format.
"*/
/*
 * Maybe it would be a good idea to have another lexer program to parse the
 * comment string... it would probably be a bit more flexible, and we could
 * maybe have different lexers to handle different kinds of markup in
 * comments in case we want to allow something other than the #{}, %{}, _{},
 * !{} stuff we have now.  At the very least, it would make all of this
 * parsing code a whole lot nicer, and probably a lot less idiosyncratic.
 */
NSString *ADFormatString(NSString *string, NSArray *italicWordsArray)
{
    ADOutputModule *outputModule = [ADOutputModule currentModule];
    NSString *startBold = [outputModule startBold];
    NSString *stopBold  = [outputModule stopBold];
    NSString *startItal = [outputModule startItalics];
    NSString *stopItal  = [outputModule stopItalics];
    NSString *boldReplace1   = [NSString stringWithFormat:@"%@$1%@",   startBold, stopBold];
    NSString *boldReplace2   = [NSString stringWithFormat:@"$1%@$2%@", startBold, stopBold];
    NSString *italicReplace1 = [NSString stringWithFormat:@"%@$1%@",   startItal, stopItal];
    NSString *italicReplace2 = [NSString stringWithFormat:@"$1%@$2%@", startItal, stopItal];

    string = [outputModule preProcessString:string];

    /* boldify text of the form #{the text} */
    string = [string stringByReplacingEveryOccurrenceOfRegex:@"#{([^}]+)}"
                        withString:boldReplace1
                        options:MiscUseMatchSubstitutions];
	
    /* italicize text of the form %{the text} */
    string = [string stringByReplacingEveryOccurrenceOfRegex:@"%{([^}]+)}"
                        withString:italicReplace1
                        options:MiscUseMatchSubstitutions];

    /* boldify words which start with a "#" in front of them */
    string = [string stringByReplacingEveryOccurrenceOfRegex:@"(\\`|[^\\\\])#((\\w|:|_)+)"
                        withString:boldReplace2
                        options:MiscUseMatchSubstitutions];

    /* italicize words which start with a "%" in front of them */
    // orig autodoc didn't have ':' as part of pattern
    string = [string stringByReplacingEveryOccurrenceOfRegex:@"(\\`|[^\\\\])%((\\w|:|_)+)"
                        withString:italicReplace2
                        options:MiscUseMatchSubstitutions];

	
    /* Reformat '\%' and '\#' and '\/' to just '%' and '#' and '/' */
    string = [string stringByReplacingEveryOccurrenceOfRegex:@"\\\\([/#%])"
                        withString:@"$1" options:MiscUseMatchSubstitutions];

    /* Format !{...} type of markup -- it will do paragraph handling and
       auto-bolding and -italicizing as well. */
    string = ADFormatFixedFontMarkup(string, italicWordsArray);

    /* Format _{...} type of markup */
    string = ADFormatListMarkup(string, outputModule);

    string = [outputModule postProcessString:string];

    return string;
}


NSString *ADIndentString(NSString *string, int numLevels, BOOL indentFirstLine)
{
    NSArray        *lineArray = [string componentsSeparatedByString:@"\n"];
    NSEnumerator   *lineEnum = [lineArray objectEnumerator];
    NSString       *line;
    NSMutableArray *outputArray = [NSMutableArray array];
    NSString       *indentString = [[ADOutputModule currentModule] valueIndentString];
    NSString       *indent = @"";
    int            i;

    for (i=0; i<numLevels; i++)
        indent = [indent stringByAppendingString:indentString];

    if (!indentFirstLine)
        [outputArray addObject:[lineEnum nextObject]];

    while (line = [lineEnum nextObject])
    {
        line = [indent stringByAppendingString:line];
        [outputArray addObject:line];
    }

    return [outputArray componentsJoinedByString:@"\n"];
}

