//
//	ADCompoundIvar.m
//	  Written by Carl Lindberg Copyright 1998 by Carl Lindberg.
//                     All rights reserved.
//		This notice may not be removed from this source code.
//
//	This header is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"License.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

#import "ADCompoundIvar.h"
#import <Foundation/NSCharacterSet.h>
#import <Foundation/NSArray.h>
#import <MiscFoundation.subproj/NSString+MiscAdditions.h>
#import "ADComment.h"
#import "ADUtilities.h"
#import "ADDefaults.h"

@implementation ADCompoundIvar

/*" Represents a "compound" ivar, i.e. a struct or a union. "*/

- (id)initWithString:(NSString *)string
{
    [super initWithString:string];

    /*
     * the "type" is really the stuff that comes after the last "}"; store
     * it in a different ivar.  Alternatively; we could store the "struct
     * XXX" part in  a different ivar and leave type alone.
     */
    variableType = @"";
    if ([type length] > 0)
        variableType = ADSquashTypeString([type substringFromEndOfString:@"}"]);
    [variableType retain];
    [self setType:nil];

    return self;
}
- (void)dealloc
{
	[subIvars release];
    [variableType release];
	[super dealloc];
}

/*"
 * Returns the string that comes between the last "}" and the variable
 * name.  Usually blank.
"*/
- (NSString *)variableType
{
    return variableType;
}

/*" Returns the array of subivars that make up the structure or union. "*/
- (NSArray *)subIvars
{
	return subIvars;
}

/*" Sets the subivars array.  This method is used by the parser. "*/
- (void)setSubIvars:(NSArray *)ivarArray
{
	[ivarArray makeObjectsPerformSelector:@selector(setParentIvar:) withObject:self];
	SETIVAR(subIvars, ivarArray);
}

/*"
 * Expects a string like "struct __flags {" and sets the type from there.
 * Intended for parser use only.
"*/
- (void)setTypeFromLeadingString:(NSString *)leadString
{
    NSString *compoundType = [leadString wordNum:0];  // "union" or "struct"
    NSString *compoundName = [leadString wordNum:1];

    if ([compoundName isEqual:@"{"])  // there is no name
        [self setType:[compoundType stringByAppendingString:@" "]];
    else
        [self setType:[NSString stringWithFormat:@"%@ %@ ", compoundType, compoundName]];
}

/*" Returns a description of the ivar. "*/
- description
{
	return [NSString stringWithFormat:@" %@ {\n%@ } %@;", type, subIvars, name];
}


- (NSString *)fullTypeWithSuffix:(BOOL)includeSuffix
{
    NSMutableArray *stringArray = [NSMutableArray array];
    int i, count = [subIvars count];

    [stringArray addObject:[[self type] stringByAppendingString:@"{"]];
    for (i=0; i<count;i ++) {
        [stringArray addObject:[[subIvars objectAtIndex:i] fullType]];
    }
    if (includeSuffix)
        [stringArray addObject:[NSString stringWithFormat:@"} %@%@%@;",
            variableType, name, posttype]];
    else
        [stringArray addObject:[NSString stringWithFormat:@"} %@", variableType]];

    return [stringArray componentsJoinedByString:@"\n"];
}

- (NSString *)fullType
{
    return [self fullTypeWithSuffix:YES];
}

/* Override to add all subivars */
- (void)addToFullIvarArray:(NSMutableArray *)array
{
    int i, count = [subIvars count];

    if (![[[self entityDescription] rawCommentString] isEqual:@"No description."])
    {
        [array addObject:self];
    }

    for (i=0; i<count; i++)
    {
        [[subIvars objectAtIndex:i] addToFullIvarArray:array];
    }
}

/*"
 * We override to build up a full type string, with all the child ivars'
 * -fullType strings, to print out in the documentation.  This is run
 * through ADFormatValueString() to format the structure as nicely as
 * possible.
"*/
- (NSString *)formattedType
{
    return ADFormatValueString([self fullTypeWithSuffix:NO]);
}

@end
