//
//	ADOutputModule.m
//		Written by Carl Lindberg Copyright 1998 by Carl Lindberg.
//				All rights reserved.
//
//		This notice may not be removed from this source code.
//
//	This header is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"License.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

#import "ADOutputModule.h"
#import <Foundation/NSUtilities.h>
#import <Foundation/NSPathUtilities.h>
#import <Foundation/NSObjCRuntime.h>
#import <Foundation/NSUserDefaults.h>
#import <Foundation/NSData.h>
#import <Foundation/NSAutoreleasePool.h>
#import <Foundation/NSArray.h>
#import <MiscFoundation.subproj/NSString+MiscAdditions.h>
#import <MiscFoundation.subproj/NSString+MiscRegex.h>
#import <MiscMerge.subproj/MiscMergeEngine.h>
#import <MiscMerge.subproj/MiscMergeTemplate.h>
#import "NSFileManager+Autodoc.h"
#import "ADDefaults.h"
#import "ADUtilities.h"
#import "ADFile.h"

static ADOutputModule *currentModule;

@interface ASCIIOutputModule : ADOutputModule
@end
@interface RTFOutputModule : ADOutputModule
@end
@interface HTMLOutputModule : ADOutputModule
@end
@interface LaTeXOutputModule : ADOutputModule
@end
@interface GSDocOutputModule : HTMLOutputModule
@end

static NSDictionary *_outputMapDictionary()
{
    return [@"{ "
                 @"rtf = RTFOutputModule; "
                 @"html = HTMLOutputModule; "
                 @"txt = ASCIIOutputModule; "
                 @"ascii = ASCIIOutputModule; "
                 @"latex = LaTeXOutputModule; "
                 @"gsdoc = GSDocOutputModule; "
            @"}"
        propertyList];
}

@implementation ADOutputModule

+ (ADOutputModule *)currentModule
{
	return currentModule;
}

+ (void)setCurrentModuleFromString:(NSString *)aString
{
	NSString *className = [_outputMapDictionary() objectForKey:[aString lowercaseString]];

	if (className == nil) className = @"RTFOutputModule";
	
	[currentModule release];
	currentModule = [[NSClassFromString(className) alloc] init];
}

/*" Returns the file extension used for generated documentation. "*/
- (NSString *)fileExtension		{ return @""; }

/*" Returns a default template name for this output type. "*/
- (NSString *)templateFilename	{ return @""; }

/*"
 * Returns an array of documentation extensions, used by the -copydocs flag
 * to find documentation files to copy to the destination directory.
"*/
- (NSArray *)documentationExtensions	{ return nil; }


/*" Returns a string to start a series of bolded text "*/
- (NSString *)startBold		{ return @""; }

/*" Returns a string to end a series of bolded text "*/
- (NSString *)stopBold		{ return @""; }

/*" Returns a string to start a series of italicized text "*/
- (NSString *)startItalics	{ return @""; }

/*" Returns a string to end a series of italicized text "*/
- (NSString *)stopItalics	{ return @""; }


/*" Return the paragraph separator, used to separate paragraphs in comments "*/
- (NSString *)paragraphSeparator { return @""; }

/*" Return a line separator, used to force a new line when formatting structures "*/
- (NSString *)lineSeparator { return @""; }


/*" Returns a string that should go before a block of list markup "*/
- (NSString *)listPrefixString 		 { return @""; 		}

/*" Returns a string that should go after a block of list markup "*/
- (NSString *)listPostfixString		 { return @""; 		}

/*"
 * Returns a string used for each list item.  $1 will be replaced with the
 * value in the left "column", and $3 will be replaced by the comment. A
 * bare $ needs to be escaped like $$.
"*/
- (NSString *)listItemReplaceString  { return @"$1 $3";	}

/*"
 * Returns the paragraph separator used for list item comments.  By
 * default, just returns the regular paragraph separator.
"*/
- (NSString *)listParagraphSeparator { return [self paragraphSeparator]; }

- (NSString *)startFixedFont { return @""; }
- (NSString *)stopFixedFont  { return @""; }

/*"
 * Return a string for indenting one level.  AutoDoc will concatenate these
 * strings to indent multiple levels.  Used to format structures and the
 * like.
"*/
- (NSString *)valueIndentString      { return @"\t"; }


/*"
 * A generic hook; called from ADFormatString() and ADFormatValueString()
 * with the entire string before markup formatting is done.  Subclasses can
 * use to escape control characters.
"*/
- (NSString *)preProcessString:(NSString *)string
{
    return string;
}
/*"
 * A generic hook; called from ADFormatString() and ADFormatValueString()
 * with the entire string after markup formatting is done.  Subclasses can
 * use to escape control characters, but be careful to not escape control
 * characters used by markup strings being returned by other methods from
 * this class.
"*/
- (NSString *)postProcessString:(NSString *)string
{
    return string;
}

/*"
 * Actually creates the MiscMergeTemplate to use for parsing.  Subclasses
 * can override to alter the settings on the template, such as the
 * delimiter strings.  Will only be called once.
"*/
- (MiscMergeTemplate *)createTemplate
{
    return [[[MiscMergeTemplate alloc] init] autorelease];
}

/*"
 * Returns the MiscMergeTemplate instance for use with generating output.
 * This method finds the file to be parsed, and uses the template from
 * -#createTemplate to parse the file.  The template is cached, and only
 * parsed once.
"*/
- (MiscMergeTemplate *)template
{
    /* We can't change output formats, so a single template instance is enough */
    static MiscMergeTemplate *template = nil;

    if (template == nil)
    {
        NSString *templateSearchFile = ADStringForDefault(ADTemplateFilename);
        NSString *templateFile;

        if (templateSearchFile == nil)
        {
            templateSearchFile = [self templateFilename];
        }

        templateFile = templateSearchFile;
        if (!ADFileExists(templateSearchFile))
        {
            NSArray *searchPath = ADTemplateSearchPath();

            ADDebugLog(3, @"Searching for %@ in %@", templateSearchFile, searchPath);

            templateFile = [[NSFileManager defaultManager]
                                                findFile:templateSearchFile inSearchPath:searchPath];
        }

        if (templateFile == nil)
        {
            ADLog(@"%@: Can't find template file %@", [self class], templateSearchFile);
            templateFile = templateSearchFile;
        }
        else
        {
            ADDebugLog(2, @"Using template file %@", ADNormalizePath(templateFile));
        }

        template = [[self createTemplate] retain];
        [template parseContentsOfFile:templateFile];
    }

    return template;
}

- (void)outputContents:(ADFile *)fileContents toFile:(NSString *)outputFilename
{
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
    MiscMergeEngine   *merger = [[[MiscMergeEngine alloc] init] autorelease];
    NSData            *outputData;
    NSString          *outputString;
    NSDictionary      *infoDict = [[NSUserDefaults standardUserDefaults]
                                              dictionaryForKey:ADInfoDictionary];

    [merger setTemplate:[self template]];
    [merger setMainObject:fileContents];
    [merger setEngineValue:infoDict forKey:@"infoDictionary"];  //user-defined values
    outputString = [merger execute:nil];
    outputData = [outputString dataUsingEncoding:[NSString defaultCStringEncoding]];
    /* GNUSTEP had problems with the copyright symbol in the RTF template */
    if (outputData == nil)
    	outputData = [outputString dataUsingEncoding:NSNEXTSTEPStringEncoding];

    outputFilename = ADPathWithExtension(outputFilename, [self fileExtension]);
    if (!ADBoolForDefault(ADRunSilently))
        ADLog(@"Generating %@", [outputFilename lastPathComponent]);
    [outputData writeToFile:outputFilename atomically:YES];
    [pool release];
}

@end

@implementation ASCIIOutputModule

- (NSString *)fileExtension          { return @"txt";  }
- (NSString *)paragraphSeparator     { return @"\n\n"; }
- (NSString *)lineSeparator          { return @"\n";   }
- (NSString *)listItemReplaceString  { return @"\t$1\t$3\n"; }
- (NSString *)listParagraphSeparator { return @"\n\n\t\t";   }

/* Until I have an ASCII template, this'll do */
- (void)outputContents:(ADFile *)fileContents toFile:(NSString *)outputFilename
{
    NSData *outputData = [[fileContents description]
                            dataUsingEncoding:[NSString defaultCStringEncoding]];

    outputFilename = ADPathWithExtension(outputFilename, [self fileExtension]);
    [outputData writeToFile:outputFilename atomically:YES];
}

@end



@implementation RTFOutputModule

- (NSString *)fileExtension    { return @"rtf"; }
- (NSString *)templateFilename { return @"RTFTemplate.rtf"; }
- (NSArray *)documentationExtensions
{
    return [NSArray arrayWithObjects:@"rtf", @"rtfd", nil];
}

/* RTFTemplate.rtf uses << ... >> as its delimiters. */
- (MiscMergeTemplate *)createTemplate
{
    MiscMergeTemplate *template = [super createTemplate];
    [template setStartDelimiter:@"<<" endDelimiter:@">>"];
    return template;
}

static NSString *EscapeRTFMarkup(NSString *string)
{
    string = [string stringByReplacingEveryOccurrenceOfString:@"\\" withString:@"_BKSLASH_"];
    string = [string stringByReplacingEveryOccurrenceOfString:@"{"  withString:@"_OPBRACE_"];
    string = [string stringByReplacingEveryOccurrenceOfString:@"}"  withString:@"_CLBRACE_"];
    return string;
}

- (NSString *)realStartBold      { return @"\\b ";  }
- (NSString *)realStopBold       { return @"\\b0 "; }
- (NSString *)realStartItalics	 { return @"\\i ";  }
- (NSString *)realStopItalics    { return @"\\i0 "; }
- (NSString *)startBold          { return @"__START_BOLD_TOKEN"; } //see below
- (NSString *)stopBold           { return @"__STOP_BOLD_TOKEN"; }
- (NSString *)startItalics       { return @"__START_ITAL_TOKEN"; } // see below
- (NSString *)stopItalics        { return @"__STOP_ITAL_TOKEN"; }
- (NSString *)startFixedFont     { return EscapeRTFMarkup(@"\\f3\\fs24 "); }
- (NSString *)stopFixedFont      { return EscapeRTFMarkup(@"\n\\f0\\fs28 "); }
- (NSString *)paragraphSeparator { return EscapeRTFMarkup(@"\\\n\\fs16 \\\n\\fs28 "); }
- (NSString *)lineSeparator      { return EscapeRTFMarkup(@"\\\n"); }

- (NSString *)listParagraphSeparator
{
    return [[self paragraphSeparator] stringByAppendingString:@"\t"];
}

- (NSString *)listItemReplaceString
{
    return EscapeRTFMarkup(@"{\\pard \\li7648 \\fi-5131 \\ri1007 \\ql \\tx7648 "
                               @"\\tx10180 \\fs16 \\\n \\fs28  $1\t$3 }\n ");
}

- (NSString *)postProcessString:(NSString *)string
{
    /*
     * Reformat curly braces so they don't prematurely end the rtf doc.
     */
    string = [string stringByReplacingEveryOccurrenceOfRegex:@"[{}\\]"
                       withString:@"\\$0"
                       options:MiscUseMatchSubstitutions];

    /* Unescape the actual RTF markup we escaped previously */
    string = [string stringByReplacingEveryOccurrenceOfString:@"_BKSLASH_" withString:@"\\"];
    string = [string stringByReplacingEveryOccurrenceOfString:@"_OPBRACE_" withString:@"{"];
    string = [string stringByReplacingEveryOccurrenceOfString:@"_CLBRACE_" withString:@"}"];

    /*
     * Because the bold and italics start strings have a space in them,
     * they can confuse later parsing if they are put in too early.  So, we
     * insert a token, and put in the actual RTF markup here.
     */
    string = [string stringByReplacingEveryOccurrenceOfString:[self startBold]
                                                   withString:[self realStartBold]];
    string = [string stringByReplacingEveryOccurrenceOfString:[self stopBold]
                                                   withString:[self realStopBold]];
    string = [string stringByReplacingEveryOccurrenceOfString:[self startItalics]
                                                   withString:[self realStartItalics]];
    string = [string stringByReplacingEveryOccurrenceOfString:[self stopItalics]
                                                   withString:[self realStopItalics]];
    return string;
}

@end

@implementation HTMLOutputModule

- (NSString *)fileExtension     { return @"html"; }
- (NSString *)templateFilename  { return @"HTMLTemplate.html"; }
- (NSArray *)documentationExtensions
{
    return [NSArray arrayWithObjects:@"htm", @"html", @"htmld", nil];
}

/* HTMLTemplate.html uses [[ ... ]] as its delimiters. */
- (MiscMergeTemplate *)createTemplate
{
    MiscMergeTemplate *template = [super createTemplate];
    [template setStartDelimiter:@"[[" endDelimiter:@"]]"];
    return template;
}

- (NSString *)startBold          { return @"<B>"; }
- (NSString *)stopBold           { return @"</B>"; }
- (NSString *)startItalics       { return @"<I>"; }
- (NSString *)stopItalics        { return @"</I>"; }
- (NSString *)startFixedFont     { return @"<CODE>"; }
- (NSString *)stopFixedFont      { return @"</CODE>"; }
- (NSString *)paragraphSeparator { return @"<P>\n\n"; }
- (NSString *)lineSeparator      { return @"<BR>\n"; }

- (NSString *)listPrefixString      { return @"<TABLE BORDER=\"3\">\n"; }
- (NSString *)listPostfixString     { return @"</TABLE>\n"; }
- (NSString *)listItemReplaceString { return @"\t<TR><TD>$1</TD><TD>$3</TD></TR>\n"; }

- (NSString *)valueIndentString     { return @"&nbsp;&nbsp;&nbsp;&nbsp;"; }

- (NSString *)preProcessString:(NSString *)string
{
    /*
     * Replace certain characters with their HTML entities.  I'm not
     * handling any other entities rignt now; only normal ascii ones.
     * 
     * Make sure the '&' one is done first.
     * 
     * This does disallow HTML markup in comments...hm.
     */
    string = [string stringByReplacingEveryOccurrenceOfString:@"&" withString:@"&amp;"];
    string = [string stringByReplacingEveryOccurrenceOfString:@"<" withString:@"&lt;"];
    string = [string stringByReplacingEveryOccurrenceOfString:@">" withString:@"&gt;"];
    string = [string stringByReplacingEveryOccurrenceOfString:@"\"" withString:@"&quot;"];

    return string;
}

@end

/* Quickie GSDoc module, which is GNUSTEP documentation (an XML format).
   We don't provide an actual template with AutoDoc, but this helps
   to produce .gsdoc files with the correct file extension at least.
   All behavior is inherited from HTML; I hope that's right.
*/
@implementation GSDocOutputModule

- (NSString *)fileExtension     { return @"gsdoc"; }
- (NSString *)templateFilename  { return @"GSDocTemplate.gsdoc"; }
- (NSArray *)documentationExtensions
{
    return [NSArray arrayWithObject:@"gsdoc"];
}

@end


@implementation LaTeXOutputModule

/* LaTeXTemplate.latex uses << ... >> as its delimiters. */
- (MiscMergeTemplate *)createTemplate
{
    MiscMergeTemplate *template = [super createTemplate];
    [template setStartDelimiter:@"<<" endDelimiter:@">>"];
    return template;
}

/*
 * Control characters appearing in -startBold, etc. need to be escaped so
 * that -postProcess can tell the difference between them and literal
 * characters.  Sometimes the literal characters can just be escaped in
 * -preProcess, but if they're autodoc markup characters (\, {, and }),
 * they can't be done that way.  Also, if the way to escape them includes
 * characters otherwise escaped in -postProcess, it gets tricky.  For
 * simplicity's sake, I'm doing all escaping in -postProcess, so we need to
 * escape *any* LaTeX control characters used in -startBold and the like.
 */
static NSString *EscapeLaTeXMarkup(NSString *string)
{
    string = [string stringByReplacingEveryOccurrenceOfString:@"\\" withString:@"!BKSLASH!"];
    string = [string stringByReplacingEveryOccurrenceOfString:@"{"  withString:@"!OPBRACE!"];
    string = [string stringByReplacingEveryOccurrenceOfString:@"}"  withString:@"!CLBRACE!"];
    string = [string stringByReplacingEveryOccurrenceOfString:@"&"  withString:@"!AMPER!"];
    string = [string stringByReplacingEveryOccurrenceOfString:@"|"  withString:@"!VERTBAR!"];
    return string;
}


- (NSString *)fileExtension      { return @"latex"; }
- (NSString *)templateFilename   { return @"LaTeXTemplate.latex"; }
- (NSArray *)documentationExtensions
{
    return [NSArray arrayWithObjects:@"latex", nil];
}

- (NSString *)startBold          { return EscapeLaTeXMarkup(@"\\textbf{"); }
- (NSString *)stopBold           { return EscapeLaTeXMarkup(@"}"); }
- (NSString *)startItalics       { return EscapeLaTeXMarkup(@"\\textit{"); }
- (NSString *)stopItalics        { return EscapeLaTeXMarkup(@"}"); }
- (NSString *)startFixedFont     { return EscapeLaTeXMarkup(@"\\texttt{"); }
- (NSString *)stopFixedFont      { return EscapeLaTeXMarkup(@"}"); }
- (NSString *)paragraphSeparator { return EscapeLaTeXMarkup(@"\\par "); }
- (NSString *)lineSeparator      { return EscapeLaTeXMarkup(@"\\\\\n"); }

- (NSString *)listPrefixString  { return EscapeLaTeXMarkup(@"\\begin{tabular}{|l|p{3in}|}\n"); }
- (NSString *)listPostfixString { return EscapeLaTeXMarkup(@"\\hline\\end{tabular}\n");	}
- (NSString *)listItemReplaceString {
    return EscapeLaTeXMarkup(@"\t\\hline $1 & \\setlength{\\parskip}{.075in} $3 \\\\\n");
}

- (NSString *)valueIndentString { return EscapeLaTeXMarkup(@"\\hspace*{2em}"); }

- (NSString *)postProcessString:(NSString *)string
{
    // This is goofy.  There are two quoting characters, \ and $, and one is used in producing
    // the other.  Thus, we have to 'move aside' one of the characters in order to avoid
    // bogus transformations.
    string = [string stringByReplacingEveryOccurrenceOfString:@"\\" withString:@"!REALBACKSLASH!"];

    // Escape some special characters like $\$$, $\&$, $\{$, $\}$, etc.
    // Math mode shouldn't really be needed (i.e. \$, \& etc should be fine),
    // but LaTeX gave some odd warnings with escaped braces not in math mode,
    // and it's easier to just do them all the same way.
    string = [string stringByReplacingEveryOccurrenceOfRegex:@"[$&%#_{}]" withString:@"$$\\$0$$"
                                                     options:MiscUseMatchSubstitutions];
    // Escape special math characters '<', '>', and '|' like $<$, $>$, $|$
    string = [string stringByReplacingEveryOccurrenceOfRegex:@"[<>|]" withString:@"$$$0$$"
                                                     options:MiscUseMatchSubstitutions];

    // Unescape actual TeX commands we escaped earlier
    string = [string stringByReplacingEveryOccurrenceOfString:@"!BKSLASH!" withString:@"\\"];
    string = [string stringByReplacingEveryOccurrenceOfString:@"!OPBRACE!" withString:@"{"];
    string = [string stringByReplacingEveryOccurrenceOfString:@"!CLBRACE!" withString:@"}"];
    string = [string stringByReplacingEveryOccurrenceOfString:@"!AMPER!"   withString:@"&"];
    string = [string stringByReplacingEveryOccurrenceOfString:@"!VERTBAR!" withString:@"|"];

    // Escape the ^, ~, and \ characters.  What a pain.
    string = [string stringByReplacingEveryOccurrenceOfString:@"~" withString:@"$\\tilde{ }$"];
    string = [string stringByReplacingEveryOccurrenceOfString:@"^" withString:@"$\\hat{ }$"];
    string = [string stringByReplacingEveryOccurrenceOfString:@"!REALBACKSLASH!"
                                                   withString:@"$\\backslash$"];

    return string;
}

@end
