/*
//	HeaderTokenizer.l
//	  Written by Carl Lindberg Copyright 1998 by Carl Lindberg.
//                     All rights reserved.
//		This notice may not be removed from this source code.
//
//	This header is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"License.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
*/	

%{

#import <Foundation/NSString.h>
#import <Foundation/NSArray.h>
#import <Foundation/NSData.h>
#import <MiscFoundation.subproj/NSString+MiscRegex.h>
#import <MiscFoundation.subproj/NSString+MiscAdditions.h>

#import "ADClass.h"
#import "ADComment.h"
#import "ADProtocol.h"
#import "ADIvar.h"
#import "ADCategory.h"
#import "ADMethod.h"
#import "ADDefaults.h"
#import "ADUtilities.h"

#define YYSTRING [NSString stringWithCString:yytext length:yyleng]
#define DATASTRING [NSString stringWithCString:[accumData bytes] length:[accumData length]]
#define ACCUMTEXT() [accumData appendBytes:yytext length:yyleng]
#define RESETDATA()	if (!accumData) accumData = [[NSMutableData alloc] init]; \
                    [accumData setLength:0]

#define RETURN_STRING(type) yylval.stringVal = YYSTRING; return (type)

typedef struct _MethodArg {
	NSString *argName;
	NSString *argType;
	NSString *varName;
} MethodArg;

#import "HeaderGrammar.m.h"

#define yylval HeaderGrammarlval

/* Use our logging function, so we can control where output goes */
#define fprintf adfprintf

static BOOL foundClass = NO;
static int numSquiggles = 0;
static int currlineno;
static NSMutableData *accumData = nil;

static void UpdateLineno(char *text, int textLength);
static void SetupHeaderTokenizer();

#define YY_USER_ACTION UpdateLineno(yytext, yyleng);
#define YY_USER_INIT   SetupHeaderTokenizer();

%}

%option debug
%option stack
%option noyy_top_state
%option never-interactive
%option noyywrap
%option noreject
%option nounput

ident				[a-zA-Z_][a-zA-Z0-9_]*
ccomment			("/*"([^*]|"*"+[^*/])*"*"+"/")
objccomment			("//"[^\n]*\n?)
objccommentblock	{objccomment}([ \t]*{objccomment})*
anycomment			{ccomment}|{objccomment}
whitechar			[ \t\r\n]
white				{whitechar}*
whiteexist			{whitechar}+
conststring			["](\\.|[^\\"])*["]
constchar			['](\\.|[^\\'])[']
statictype			(\([^/:;@+-]*\))
methodstart			^{white}[+-]{white}{statictype}?
methodargument		":"{white}{statictype}?{white}{ident}
extern				(extern|{ident}_(EXTERN|EXPORT))
staticinline        (static{whiteexist}(inline|__inline__|INLINE)|{ident}_STATIC_INLINE)

%s CLASS
%s METHOD
%s VARIABLES
%s VARSTRUCT
%x INTYPEDEF
%x INENUM
%x SQUIGGLES
%x EXTERNSTRUCT

%%

{ccomment}			|
{objccommentblock}	{
						NSString *commentString = YYSTRING;
						
						if (YYSTATE == INITIAL && ADIsCopyrightComment(commentString))
						{
							yylval.stringVal = commentString;
							return COPYRIGHTCOMMENT;
						}

						if (ADIsAutodocComment(commentString))
						{
							yylval.stringVal = commentString;
							return AUTODOCCOMMENT;
						}
					}

<*>"#"{white}(import|include){white}"<"[^>]+">"	;
<*>"#"{white}(import|include){white}"\""[^\"]+"\""	;
<*>"#"{white}(ifdef|ifndef|if|else|elif|endif|undef|warning|error|pragma)(\\\n|.)*\n	;

@class{white}{ident}({white},{white}{ident})*{white}";"	;

	/* Ignore 'extern "C"' type things. */
{extern}{white}{conststring}	;

    /*
     * BUG: global variables that are function pointers (for example,
     * "extern int (*someFunc)(int, float);") end up matching the FUNCTION
     * rule rather than the GLOBALVAR rule.  This is a difficult one to fix.
     * The following is an attempt, but it might break actual functions
     * that have function pointer arguments (or return values).
     */
{extern}{whiteexist}[^;("{]*[(]{white}[*]{ident}[^;]+")"{white}";"	{
							if (!foundClass && YYSTATE == INITIAL)
							{
								RETURN_STRING(GLOBALVAR);
							}
						}

{extern}{whiteexist}[^;"{][^;{]*")"{white}";"	{
							if (!foundClass && YYSTATE == INITIAL)
							{
								RETURN_STRING(FUNCTION);
							}
						}

{extern}{whiteexist}[^;"{][^;{]*";"			{
							if (!foundClass && YYSTATE == INITIAL)
							{
								RETURN_STRING(GLOBALVAR);
							}
						}

          /* A global variable that's typed as a struct or union, ugh. */
{extern}{whiteexist}[^;"{]*[{]   {
						    yy_push_state(EXTERNSTRUCT);
                            RESETDATA();
                            ACCUMTEXT();
						    numSquiggles++;
                        }

<EXTERNSTRUCT>{anycomment}	;
<EXTERNSTRUCT>[^{}/;]+	    { ACCUMTEXT(); }
<EXTERNSTRUCT>"/"			{ ACCUMTEXT(); } // A slash that's not the start of a comment
<EXTERNSTRUCT>"{"			{ ACCUMTEXT(); numSquiggles++; }
<EXTERNSTRUCT>"}"			{ ACCUMTEXT(); numSquiggles--; }
<EXTERNSTRUCT>";"           { ACCUMTEXT(); 
                                if (numSquiggles == 0)
							    {
								    yy_pop_state();
								    yylval.stringVal = DATASTRING;
								    if (!foundClass && YYSTATE == INITIAL) return GLOBALVAR;
							    }
						    }


{staticinline}{whiteexist}[^{]+"{"    {

						    int currState = YYSTATE;
						    yy_push_state(SQUIGGLES);
						    numSquiggles++;
							if (!foundClass && currState == INITIAL)
							{
							    RETURN_STRING(INLINEFUNCTION);
							}
						}

<SQUIGGLES>{anycomment}	;
<SQUIGGLES>{conststring}	;
<SQUIGGLES>{constchar}	;
<SQUIGGLES>[^{}/"']+		;
<SQUIGGLES>"/"			;  // A slash that's not the start of a comment
<SQUIGGLES>"{"			{	numSquiggles++; }
<SQUIGGLES>"}"			{	numSquiggles--; 
							if (numSquiggles == 0)
							{
								yy_pop_state();
							}
						}

(static|inline){whiteexist}[^;{]+";"			;

"#"{white}define{white}{ident}"("(\\\n|.)+  {
							if (!foundClass && YYSTATE == INITIAL)
							{
								RETURN_STRING(MACRO); 
							}
						}

"#"{white}define{white}{ident}(\\\n|.)+  {
							if (!foundClass && YYSTATE == INITIAL)
							{
								RETURN_STRING(CONSTANT); 
							}
						}

typedef{whiteexist}	{	yy_push_state(INTYPEDEF);
								RESETDATA();
								ACCUMTEXT();
							}
<INTYPEDEF>"{"			{	ACCUMTEXT(); numSquiggles++;	}
<INTYPEDEF>"}"			{	ACCUMTEXT(); numSquiggles--;	}
<INTYPEDEF>{anycomment}	;
<INTYPEDEF>[^;{}/#]+	{	ACCUMTEXT(); }
<INTYPEDEF>[/#]+		{	ACCUMTEXT(); }
<INTYPEDEF>";"			{	ACCUMTEXT();
							if (numSquiggles == 0)
							{
								yy_pop_state();
								yylval.stringVal = DATASTRING;
								if (!foundClass && YYSTATE == INITIAL) return TYPEDEF;
							}
						}

obsolete{whiteexist}	;	/* A "keyword" that occasionally pops up in NeXT's headers */

enum{whiteexist}	{
							yy_push_state(INENUM);
							RESETDATA();
							ACCUMTEXT();
						}
<INENUM>{constchar}     {	ACCUMTEXT(); }
<INENUM>"{"				{	ACCUMTEXT(); numSquiggles++;	}
<INENUM>"}"				{	ACCUMTEXT(); numSquiggles--;	}
<INENUM>{anycomment}	;
<INENUM>[^;{}'/#]+		{	ACCUMTEXT(); }
<INENUM>['/#]+			{	ACCUMTEXT(); }
<INENUM>";"				{	ACCUMTEXT();
							if (numSquiggles == 0)
							{
								yy_pop_state();
								yylval.stringVal = DATASTRING;
								if (!foundClass && YYSTATE == INITIAL) return ENUM;
							}
						}

@interface				{	BEGIN(CLASS); return INTERFACE; }
@protocol				{	BEGIN(CLASS); return PROTOCOL; }

<CLASS>[:()<>,]			{	return yytext[0]; }

<CLASS>@end				{	BEGIN(INITIAL); return END;	}

<CLASS>"{"				{	BEGIN(VARIABLES); return yytext[0]; }

<VARIABLES>"}"			{	BEGIN(CLASS); return yytext[0]; }

						/* Maybe we shouldn't document @private ivars at all */
<VARIABLES>(@public|@protected|@private)	;

<VARIABLES,VARSTRUCT>[^ \t\r\n;/{}@#][^;{}]+";"	        { RETURN_STRING(VARDECL); }

<VARIABLES,VARSTRUCT>struct{white}({ident}{white})?"{"  {
                            yy_push_state(VARSTRUCT);
                            RETURN_STRING(BEGINSTRUCT);
                        }
<VARIABLES,VARSTRUCT>union{white}({ident}{white})?"{"	{
                            yy_push_state(VARSTRUCT);
                            RETURN_STRING(BEGINSTRUCT);
                        }
<VARSTRUCT>"}"{white}[^-+;{}/@"]*{ident}[^;{}]*";"	 { yy_pop_state();  RETURN_STRING(ENDSTRUCT); }


{ident}					{	if (YYSTATE != INITIAL)
							{
								RETURN_STRING(IDENT);
							}
						}

<CLASS>[+-]				{	BEGIN(METHOD); yylval.charVal = yytext[0];	return METHODSTART; }

<METHOD>[:,]			{	return yytext[0]; }

<METHOD>{statictype}	{	RETURN_STRING(STATICTYPE);	}
<METHOD>"..."			{	RETURN_STRING(ELLIPSIS); }

<METHOD>";"				{	BEGIN(CLASS);	return yytext[0]; }

{whiteexist}				;
.|\n						;


%%

static void UpdateLineno(char *text, int textLength)
{
	int i;

	for (i=0; i<textLength; i++)
		if (text[i] == '\n') currlineno++;
}

static void SetupHeaderTokenizer()
{
#ifdef FLEX_DEBUG
	yy_flex_debug = (ADIntForDefault(ADDebugLevel) >= 4);
#endif
}

void ADSetHeaderFile(FILE *file)
{
	foundClass = NO;
	numSquiggles = 0;
	currlineno = 1;
    BEGIN(INITIAL);
	yyin = file;
}

int ADHeaderLineNumber()
{
	return currlineno;
}
